import sys
import os
import string
# modify import path instead of adding a __init__.py as I want to keep 
# a the Sulley install as unmodified as possible to facilitate easy updating 
# if there are changes to it. 
# Props to Pedram Amini and Aaron Portnoy for creating such a great framework
sys.path.append(''.join([os.getcwd(), '/sulley']))

from random import Random
from protocol_logic.sip_utilities import SIPCrashDetector
from protocol_logic.sip_utilities import SIPInviteCanceler
from misc.utilities import Logger

from socket import *        
from sulley import *

class AbstractFuzzer:    
    def __init__(self, audit_folder, proto, host, port, crash_detection, skip=0, start_command='', stop_command='TERMINATE_PID', \
                 procmon_port=26002, restart_interval=0, log_function=None, max_len=8192):
        '''
        Abstract class to be implemented by the various fuzzers. Does not
        provide any 'fuzzing' implementation, just utility functions.
        Basically an abstraction around a Sulley fuzzer to make things a
        bit nicer for integration and management of a variety of requests.

        @type audit_folder: String
        @param audit_folder: The folder to which session information is stored
        @type proto: String
        @param proto: (Optional, def=udp) The protocol to use. TCP or UDP
        @type host: String
        @param host: The host to target
        @type port: Integer
        @param port: The port to target
        @type crash_detection: Integer
        @param crash_detection: Controls crash detection settings. 0 -
        Disabled, 1 - Logging enabled, 2 - Logging enabled and fuzzer paused, 3 - use process monitoring scripts in sulley/ subdirectory
        @type skip: Number of tests to skip
        @param skip: Integer
        @type start_command: String
        @param start_command: The command used on the target system to start the target program
        @type stop_command: String
        @param stop_command: (Optional, def=TERMINATE_PID)The command used on the target system to stop the target program
        @type procmon_port: Integer
        @param procmon_port: (Optional, def=26002)The port the remote process monitoring script is listening on
        @type log_function: Function
        @param log_functin: A function taking a string and a log level that the class will use to log data
        @type restart_interval: Integer
        @param restart_interval: The interval with which to restart the target application. Only apples with crash_detection == 3
        @type max_len: Integer
        @param max_len: The maximum length of fuzz strings to use
        '''
        
        self.audit_folder = audit_folder
        self.proto = proto.lower()
        self.host = host
        self.port = port
        self.crash_detection = crash_detection
        self.skip = skip

        # This actually causes me physical discomfort!
        # The reason I've had to resort to it is due to the way Sulley is set up
        # When you import your protocol mapping the strings are created so to allow
        # the user to specify string length I have had to move that string generation
        # to another method, call it manually and then import the protocol mapping
        primitives.gen_strings(max_len)
        from requests import sip

        self.sess = sessions.session(session_filename=self.audit_folder + \
        '/sulley.session', skip=self.skip, proto=self.proto, audit_folder=self.audit_folder, restart_interval=restart_interval)
        
        self.target = sessions.target(self.host, self.port)
        self.start_command = start_command
        self.stop_command = stop_command
        self.procmon_port = procmon_port
        self.using_GUI = False
        # get default logging method. This can be overridden by a GUI interface
        if log_function:
            self.log = log_function
        else:
            self.log = Logger().log

        self.sess.log = self.log
        
        if self.crash_detection == 3:
            self.target.procmon = pedrpc.client(self.host, self.procmon_port)
            self.target.procmon_options = \
                       {
                           "stop_commands" : [self.stop_command],
                           "start_commands" : [self.start_command],
                        }
        # the following two variables are lists of functions that will be
        # registered as pre/post send methods with Sulley
        self.pre_send_functions = []
        self.post_send_functions = []
        self.sess.pre_send = self.pre_send
        self.sess.post_send = self.post_send
        self.sess.close_sock = self.close_sock

    def close_sock(self, sock):
        '''
        Because the INVITE canceller needs sockets to survive longer than
        sulley would allow, the responsbibility for closing sockets had to be
        moved out here

        @type sock: Socket
        @param sock: The socket to be closed
        '''
        
        sock.close()
        
    def fuzz(self):
        pass

    def pre_send(self, sock):
        '''
        Iterates through the list of registered pre send functions and calls
        them

        @type sock: Socket
        @param sock: The socket used to send the last test case
        '''
        
        for function in self.pre_send_functions:
            function(sock)

    def post_send(self, sock):
        '''
        Iterates through the list of registered post send functions and calls
        them

        @type sock: Socket
        @param sock: The socket used to send the last test case
        '''
        
        for function in self.post_send_functions:
            function(sock)
        
    def detect_crash(self):
        '''
        Interface method to be implemented by child classes
        '''
        pass

    def empty_func(self):
        '''
        Procedure with no functionality. Useful for removing parts of classes in Sulley
        that we don't need/want without breaking things
        '''
        pass

################################################################################

class AbstractSIPFuzzer(AbstractFuzzer):
    def __init__(self, audit_file, proto, host, port, crash_detection, skip=0, start_command='TERMINATE_PID', stop_command='', \
                 procmon_port=26002, restart_interval=0, log_function=None, max_len=8192):
        '''
        Abstract parent class for all SIP based fuzzers. Simply updates the default
        fuzzer to include SIP specific crash detection
        '''
        
        AbstractFuzzer.__init__(self, audit_file, proto, host, port, 
            crash_detection, skip, start_command, stop_command, procmon_port,
            restart_interval, log_function, max_len)

        if self.crash_detection > 0 and self.crash_detection < 3:
            self.post_send_functions.append(self.detect_crash)
            self.crash_detector = SIPCrashDetector(self.host, self.port)
            self.crash_detector.log = self.log
    
    def generate_unique_attributes(self, session, node, edge, sock):
        '''
        In SIP a number of fields need to be unique/random e.g branch. Without
        this some implementations freak out and don't function in a normal
        fashion, thus artificially interfering with the fuzzing process. The
        SIP rfc specifies cryptographically random values should be used but
        for our purposes the pseudo random values generated should do.
       
        @rtype: String
        @return: Returns the SIP message with the fields requiring random
        identifiers filled in correctly when possible i.e if they're not being
        fuzzed
        '''

        data = node.render()
        # branch
        x = ''.join(Random().sample(string.ascii_lowercase+string.digits, 32)) 
        data = data.replace('somebranchvalue', x)
        # from tag
        x = ''.join(Random().sample(string.ascii_lowercase+string.digits, 32)) 
        data = data.replace('somefromtagvalue', x)
        # call-id tag
        x = ''.join(Random().sample(string.ascii_lowercase+string.digits, 32)) 
        data = data.replace('somecallidvalue', x)
       
        return data
        
    def detect_crash(self, sock):
        '''
        This method is registered as the post-send callback with Sulley and
        attempts to detect crashes by sending an OPTIONS message to the target
        and waiting for a response. If no response is received it will notify
        the user and log the last request sent.
        WARNING: Ekiga gives a lot of false positives to this form of detection

        @type sock: Socket
        @param sock: The socket used to send the last test case. Unused in this
        method but necessary as a parameter as it is passed to all post_send
        methods

        @todo: Add support for automatically restarting the target program
        possibly by modifying 'process_monitor.py' from Sulley
        @todo: Think of a way to make this work with ekiga which doesn't
        respond to OPTIONS messages if it is processing a call
        '''
      
        if not self.crash_detector.is_responding():
            self.log('[*] The target program has stopped responding')
            crash_log_name = self.audit_folder + '/' + \
                str(self.sess.fuzz_node.id) + '_' + \
                str(self.sess.total_mutant_index) + '.crashlog'
            crash_log = open(crash_log_name, 'w')
            crash_log.write(self.sess.fuzz_node.sent_data)
            crash_log.close()
            self.log('[*] Fuzz request logged to ' + crash_log_name)
            if self.crash_detection == 2:
                if self.using_GUI:
                    # the GUI will hopefully have filled in this method
                    self.pause_GUI()
                else:
                    raw_input('Press the any key to continue\n')

################################################################################

class AbstractSIPInviteFuzzer(AbstractSIPFuzzer):
    def __init__(self, audit_file, proto, host, port, crash_detection, skip=0, start_command='TERMINATE_PID', stop_command='', \
                 procmon_port=26002,  restart_interval=0, log_function=None, max_len=8192):
        '''
        Abstract parent class for all SIP INVITE  based fuzzers. Updates the default
        SIP fuzzer to include CANCEL messages for the sent INVITES to prevent
        certain clients dying from INVITE floods (not really that interested in
        finding those ;) )
        '''
        
        AbstractSIPFuzzer.__init__(self, audit_file, proto, host, port,
            crash_detection, skip, start_command, stop_command, procmon_port,
            restart_interval,  log_function, max_len)
        self.invite_canceler = SIPInviteCanceler(self.host, self.port, 2.0)
        self.invite_canceler.log = self.log
        self.post_send_functions.append(self.cancel)

        # any fuzzers that need to handle the closing of the post_send socket
        # themselves should include the following line
        self.sess.close_sock = self.empty_func
    
    def cancel(self, sock):
        '''
        Calls an external class to cancel the last INVITE sent

        @type sock: Socket
        @param sock: The socket used to send the last test case
        '''

        self.invite_canceler.cancel(sock, self.sess.fuzz_node.sent_data)
            
################################################################################
